# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import os
import time

from heatclient.v1 import client as heatclient
from heatclient import exc
import mock
import yaml

import tobiko
from tobiko.openstack import heat
from tobiko.openstack import keystone
from tobiko.tests.unit import openstack


class MyStack(heat.HeatStackFixture):
    pass


class MyStackWithStackName(heat.HeatStackFixture):
    stack_name = 'stack.name.from.class'


class MyStackWithParameters(heat.HeatStackFixture):
    parameters = {'param': 'from-class'}


class MyStackWithTemplate(heat.HeatStackFixture):
    template = {'template': 'from-class'}


class MyStackWithWaitInterval(heat.HeatStackFixture):
    wait_interval = 10


class MyTemplateFixture(tobiko.SharedFixture):

    _template = {'template': 'from-class'}
    template = None

    def __init__(self, template=None):
        super(MyTemplateFixture, self).__init__()
        if template:
            self._template = template

    def setup_fixture(self):
        self.template = heat.HeatTemplate.from_dict(self._template)


class HeatStackFixtureTest(openstack.OpenstackTest):

    def test_init(self, fixture_class=MyStack, stack_name=None, template=None,
                  parameters=None, wait_interval=None, client=None):
        stack = fixture_class(stack_name=stack_name, template=template,
                              parameters=parameters,
                              wait_interval=wait_interval, client=client)

        self.assertEqual(stack_name or fixture_class.stack_name or
                         tobiko.get_fixture_name(stack),
                         stack.stack_name)

        if tobiko.is_fixture(template):
            self.assertIsNone(stack.template)
            self.assertIs(template, stack.template_fixture)
        elif isinstance(template, collections.Mapping):
            self.assertEqual(
                heat.HeatTemplate.from_dict(template=template),
                stack.template)
            self.assertIsNone(stack.template_fixture)
        elif template:
            self.assertIs(template, stack.template)
            self.assertIsNone(stack.template_fixture)
        else:
            self.assertIsNone(stack.template)
            self.assertIsNone(stack.template_fixture)

        self.assertIs(fixture_class.parameters, stack.parameters)

        self.assertEqual(wait_interval or fixture_class.wait_interval,
                         stack.wait_interval)

        if tobiko.is_fixture(client):
            self.assertIsNone(stack.client)
            self.assertIs(client, stack.client_fixture)
        elif client:
            self.assertIs(client, stack.client)
            self.assertIsNone(stack.client_fixture)
        else:
            self.assertIsNone(stack.client)
            self.assertIsNone(stack.client_fixture)

    def test_init_with_stack_name(self):
        self.test_init(stack_name='my-stack-name')

    def test_init_with_stack_name_from_class(self):
        self.test_init(fixture_class=MyStackWithStackName)

    def test_init_with_template(self):
        self.test_init(template={'some': 'template'})

    def test_init_with_template_fixture(self):
        self.test_init(template=MyTemplateFixture())

    def test_init_with_parameters(self):
        self.test_init(parameters={'my': 'value'})

    def test_init_with_parameters_from_class(self):
        self.test_init(fixture_class=MyStackWithParameters)

    def test_init_with_wait_interval(self):
        self.test_init(wait_interval=20)

    def test_init_with_wait_interval_from_class(self):
        self.test_init(fixture_class=MyStackWithWaitInterval)

    def test_init_with_client(self):
        session = keystone.get_keystone_session()
        self.test_init(client=heatclient.Client(session=session))

    def test_init_with_client_fixture(self):
        self.test_init(client=heat.HeatClientFixture())

    def test_setup(self, fixture_class=MyStack, stack_name=None,
                   template=None, parameters=None, wait_interval=None,
                   stacks=None, create_conflict=False,
                   call_create=True, call_delete=False, call_sleep=False):
        from tobiko.openstack.heat import _client
        from tobiko.openstack.heat import _template

        client = mock.MagicMock(specs=heatclient.Client)
        get_heat_client = self.patch(_client, 'get_heat_client',
                                     return_value=client)

        stacks = stacks or [
            exc.HTTPNotFound,
            mock_stack('CREATE_IN_PROGRESS')]
        client.stacks.get.side_effect = stacks

        if create_conflict:
            client.stacks.create.side_effect = exc.HTTPConflict
        else:
            client.stacks.create.return_value = {
                'stack': {'id': '<stack-id>'}}

        sleep = self.patch(time, 'sleep')
        stack = fixture_class(stack_name=stack_name, parameters=parameters,
                              template=template, wait_interval=wait_interval)

        default_template = heat.HeatTemplate.from_dict(
            {'default': 'template'})
        get_heat_template = self.patch(_template, 'get_heat_template',
                                       return_value=default_template)

        stack.setUp()

        self.assertIs(client, stack.client)
        self.assertEqual(wait_interval or fixture_class.wait_interval,
                         stack.wait_interval)
        get_heat_client.assert_called_once_with()
        client.stacks.get.assert_has_calls([mock.call(stack.stack_name,
                                                      resolve_outputs=False)])

        if call_delete:
            client.stacks.delete.assert_called_once_with(stack.stack_id)
        else:
            client.stacks.delete.assert_not_called()

        self.assertEqual(parameters or fixture_class.parameters or {},
                         stack.parameters)
        if call_create:
            client.stacks.create.assert_called_once_with(
                parameters=stack.parameters, stack_name=stack.stack_name,
                template=yaml.safe_dump(stack.template.template))
        else:
            client.stacks.create.assert_not_called()

        if call_sleep:
            sleep.assert_has_calls([mock.call(stack.wait_interval)])

        if tobiko.is_fixture(template):
            self.assertIs(tobiko.get_fixture(template).template,
                          stack.template)
        elif isinstance(template, collections.Mapping):
            self.assertEqual(heat.HeatTemplate.from_dict(template),
                             stack.template)
        elif isinstance(template, heat.HeatTemplate):
            self.assertIs(template, stack.template)
        elif not template:
            if fixture_class.template:
                self.assertEqual(
                    heat.HeatTemplate.from_dict(fixture_class.template),
                    stack.template)
            else:
                self.assertEqual(default_template, stack.template)
        else:
            self.fail("Unsupported template type: {!r}".format(template))

        if template or fixture_class.template:
            get_heat_template.assert_not_called()
        else:
            template_file_name = stack.stack_name.rsplit('.', 1)[-1] + '.yaml'
            get_heat_template.assert_called_once_with(
                template_file=template_file_name,
                template_dirs=[os.path.dirname(__file__)])

    def test_setup_with_stack_name(self):
        self.test_setup(stack_name='my-stack-name')

    def test_setup_with_stack_name_from_class(self):
        self.test_setup(fixture_class=MyStackWithStackName)

    def test_setup_with_template(self):
        self.test_setup(template={'other': 'template'})

    def test_setup_with_template_from_class(self):
        self.test_setup(fixture_class=MyStackWithTemplate)

    def test_setup_with_template_fixture(self):
        self.test_setup(template=MyTemplateFixture(template={'template':
                                                             'from-fixture'}))

    def test_setup_with_template_fixture_type(self):
        self.test_setup(template=MyTemplateFixture)

    def test_setup_with_parameters(self):
        self.test_setup(parameters={'from_init': True})

    def test_setup_with_parameters_from_class(self):
        self.test_setup(fixture_class=MyStackWithParameters)

    def test_setup_with_wait_interval(self):
        self.test_setup(wait_interval=10.)

    def test_setup_with_wait_interval_from_class(self):
        self.test_setup(fixture_class=MyStackWithWaitInterval)

    def test_setup_when_delete_completed(self):
        self.test_setup(stacks=[mock_stack('DELETE_COMPLETE'),
                                mock_stack('CREATE_IN_PROGRESS')])

    def test_setup_when_delete_failed(self):
        self.test_setup(stacks=[mock_stack('DELETE_FAILED'),
                                mock_stack('DELETE_IN_PROGRESS'),
                                mock_stack('DELETE_COMPLETE'),
                                mock_stack('CREATE_IN_PROGRESS')],
                        call_delete=True, call_sleep=True)

    def test_setup_when_delete_failed_fast_delete(self):
        self.test_setup(stacks=[mock_stack('DELETE_FAILED'),
                                mock_stack('DELETE_COMPLETE'),
                                mock_stack('CREATE_IN_PROGRESS')],
                        call_delete=True)

    def test_setup_when_create_complete(self):
        self.test_setup(stacks=[mock_stack('CREATE_COMPLETE')],
                        call_create=False)

    def test_setup_when_create_failed(self):
        self.test_setup(stacks=[mock_stack('CREATE_FAILED'),
                                mock_stack('DELETE_IN_PROGRESS'),
                                mock_stack('DELETE_COMPLETE'),
                                mock_stack('CREATE_IN_PROGRESS')],
                        call_delete=True, call_sleep=True)

    def test_setup_when_create_failed_fast_delete(self):
        self.test_setup(stacks=[mock_stack('CREATE_FAILED'),
                                mock_stack('DELETE_COMPLETE'),
                                mock_stack('CREATE_IN_PROGRESS')],
                        call_delete=True)

    def test_setup_when_create_in_progress(self):
        self.test_setup(stacks=[mock_stack('CREATE_IN_PROGRESS')],
                        call_create=False)

    def test_setup_when_delete_in_progress_then_complete(self):
        self.test_setup(stacks=[mock_stack('DELETE_IN_PROGRESS'),
                                mock_stack('DELETE_COMPLETE'),
                                mock_stack('CREATE_IN_PROGRESS')],
                        call_sleep=True)

    def test_setup_when_delete_in_progress_then_failed(self):
        self.test_setup(stacks=[mock_stack('DELETE_IN_PROGRESS'),
                                mock_stack('DELETE_FAILED'),
                                mock_stack('DELETE_COMPLETE'),
                                mock_stack('CREATE_IN_PROGRESS')],
                        call_sleep=True, call_delete=True)

    def test_setup_when_create_conflict(self):
        self.test_setup(create_conflict=True)

    def test_cleanup(self):
        client = mock.MagicMock(specs=heatclient.Client)
        stack = MyStack(client=client)
        stack.cleanUp()
        client.stacks.delete.assert_called_once_with(stack.stack_name)

    def test_get_outputs(self):
        stack = mock_stack(status='CREATE_COMPLETE',
                           outputs=[{'output_key': 'key1',
                                     'output_value': 'value1'},
                                    {'output_key': 'key2',
                                     'output_value': 'value2'}])
        client = mock.MagicMock(specs=heatclient.Client)
        client.stacks.get.return_value = stack
        stack_fixture = MyStack(client=client)

        outputs = stack_fixture.get_outputs()

        client.stacks.get.assert_called_once_with(stack_fixture.stack_name,
                                                  resolve_outputs=True)
        self.assertEqual('value1', outputs.key1)
        self.assertEqual('value2', outputs.key2)


def mock_stack(status, stack_id='<stack-id>', outputs=None):
    return mock.MagicMock(stack_status=status,
                          id=stack_id,
                          outputs=outputs or [])



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack.keystone import _credentials
from tobiko.openstack.keystone import _session

keystone_credentials = _credentials.keystone_credentials
default_keystone_credentials = _credentials.default_keystone_credentials
KeystoneCredentials = _credentials.KeystoneCredentials
InvalidKeystoneCredentials = _credentials.InvalidKeystoneCredentials

KeystoneSessionFixture = _session.KeystoneSessionFixture
KeystoneSessionManager = _session.KeystoneSessionManager
get_keystone_session = _session.get_keystone_session



# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from collections import namedtuple
import os

from ansible.executor import playbook_executor
from ansible.inventory.manager import InventoryManager
from ansible.parsing.dataloader import DataLoader
from ansible.vars.manager import VariableManager
from oslo_log import log

from tobiko import config
from tobiko.openstack import keystone


LOG = log.getLogger(__name__)
CONF = config.CONF


TEMPLATE_SUFFIX = ".yaml"


class AnsibleManager(object):
    """Manages Ansible entities."""

    def __init__(self, playbooks_dir):
        self.playbooks_dir = playbooks_dir
        self.loader = DataLoader()
        self.inventory = InventoryManager(loader=self.loader,
                                          sources='localhost,')
        self.variable_manager = VariableManager(loader=self.loader,
                                                inventory=self.inventory)
        self.options = self.get_options()
        self.passwords = dict(vault_pass='secret')

    def get_playbooks_names(self, strip_suffix=False):
        """Returns a list of all the files in playbooks dir."""
        playbooks = []
        for (_, _, files) in os.walk(self.playbooks_dir):
            playbooks.extend(files)
        if strip_suffix:
            playbooks = [
                f[:-len(TEMPLATE_SUFFIX)] for f in playbooks]
        return playbooks

    def get_options(self):
        """Returns namedtuple of Ansible options."""
        Options = namedtuple('Options', ['connection', 'module_path',
                                         'forks', 'become', 'become_method',
                                         'become_user', 'check', 'diff',
                                         'listhosts', 'listtasks',
                                         'listtags', 'syntax'])

        options = Options(connection='local', module_path=['/to/mymodules'],
                          forks=10, become=None, become_method=None,
                          become_user=None, check=False, diff=False,
                          listhosts=False,
                          listtasks=False, listtags=False, syntax=False)

        return options

    def run_playbook(self, playbook, mode='create'):
        """Executes given playbook."""
        playbook_path = self.playbooks_dir + '/' + playbook

        credentials = keystone.default_keystone_credentials()
        extra_vars = {'mode': mode,
                      'auth_url': credentials.auth_url,
                      'username': credentials.username,
                      'project_name': credentials.project_name,
                      'password': credentials.project_name.password,
                      'image': CONF.tobiko.nova.image,
                      'flavor': CONF.tobiko.nova.flavor}

        self.variable_manager.extra_vars = extra_vars

        pb_executor = playbook_executor.PlaybookExecutor(
            playbooks=[playbook_path],
            inventory=self.inventory,
            variable_manager=self.variable_manager,
            loader=self.loader,
            options=self.options,
            passwords=self.passwords)

        pb_executor.run()



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os

import jinja2

from oslo_log import log

from tobiko.fault import constants as fault_const
from tobiko.common.utils import file as file_utils
from tobiko.openstack import nova


LOG = log.getLogger(__name__)


class FaultConfig(object):
    """Responsible for managing faults configuration."""

    DEFAULT_CONF_PATH = os.path.expanduser('~/.config/openstack')
    DEFAULT_CONF_NAME = "os-faults.yml"
    DEFAULT_CONF_FILE = os.path.join(DEFAULT_CONF_PATH, DEFAULT_CONF_NAME)

    def __init__(self, conf_file):
        self.templates_dir = os.path.join(os.path.dirname(__file__),
                                          'templates')
        if conf_file:
            self.conf_file = conf_file
        else:
            conf_file = self.DEFAULT_CONF_FILE
            if os.path.isfile(conf_file):
                self.conf_file = conf_file
            else:
                self.conf_file = self.generate_config_file()

    def generate_config_file(self):
        """Generates os-faults configuration file."""
        LOG.info("Generating os-fault configuration file.")
        file_utils.makedirs(self.DEFAULT_CONF_PATH)
        rendered_conf = self.get_rendered_configuration()
        with open(self.DEFAULT_CONF_FILE, "w") as f:
            f.write(rendered_conf)
        return self.DEFAULT_CONF_FILE

    def get_rendered_configuration(self):
        """Returns rendered os-fault configuration file."""
        j2_env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(self.templates_dir),
            trim_blocks=True)
        template = j2_env.get_template('os-faults.yml.j2')
        nodes = self.get_nodes()
        return template.render(nodes=nodes,
                               services=fault_const.SERVICES,
                               containers=fault_const.CONTAINERS)

    def get_nodes(self):
        """Returns a list of dictionaries with nodes name and address."""
        client = nova.get_nova_client()
        return [{'name': server.name,
                 'address': server.addresses['ctlplane'][0]['addr']}
                for server in client.servers.list()]



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko
from tobiko import config
from tobiko.shell import ping
from tobiko.shell import ssh
from tobiko.shell import sh
from tobiko.openstack import heat
from tobiko.openstack import neutron
from tobiko.openstack import stacks
from tobiko.tests import base
from tobiko.tests.scenario.neutron import _stacks


CONF = config.CONF


class FloatingIPFixture(heat.HeatStackFixture):
    """Heat stack for testing a floating IP instance
    """

    #: Heat template file
    template = _stacks.heat_template_file('floating_ip.yaml')

    #: Floating IP network where the Neutron floating IP is created
    floating_network = CONF.tobiko.neutron.floating_network

    #: Glance image used to create a Nova server instance
    image = CONF.tobiko.nova.image

    #: Nova flavor used to create a Nova server instance
    flavor = CONF.tobiko.nova.flavor

    #: Username used to login to Nova server instance
    username = CONF.tobiko.nova.username

    @property
    def internal_network(self):
        """Internal network where the Nova server instance is connected

        """
        return self.internal_network_stack.outputs.network_id

    # --- required fixtures ---

    #: Heat stack for creating internal network with a router to floating
    #:  network
    internal_network_stack = tobiko.required_setup_fixture(
        _stacks.InternalNetworkFixture)

    # --- class parameters ---
    #: Whenever port security on internal network is enable
    key_pair_stack = tobiko.required_setup_fixture(
        stacks.NovaKeyPairStackFixture)
    port_security_enabled = False

    #: Security groups to be associated to network ports
    security_groups = []

    def setup_parameters(self):
        """Setup template parameters

        """
        super(FloatingIPFixture, self).setup_parameters()
        if self.port_security_enabled:
            self.setup_port_security()

    @neutron.skip_if_missing_networking_extensions('port-security')
    def setup_port_security(self):
        """Setup port security template parameters

        """
        self.parameters.update(
            port_security_enabled=self.port_security_enabled,
            security_groups=self.security_groups or [])

    @property
    def key_name(self):
        return self.key_pair_stack.outputs.key_name

    @property
    def server_name(self):
        return self.outputs.server_name

    @property
    def floating_ip_address(self):
        return self.outputs.floating_ip_address

    @property
    def ssh_client(self):
        return ssh.ssh_client(host=self.floating_ip_address,
                              username=self.username)

    @property
    def ssh_command(self):
        return ssh.ssh_command(host=self.floating_ip_address,
                               username=self.username)


class FloatingIPTest(base.TobikoTest):
    """Tests connectivity to Nova instances via floating IPs"""

    #: Resources stack with floating IP and Nova server
    floating_ip_stack = tobiko.required_setup_fixture(FloatingIPFixture)

    @property
    def floating_ip_address(self):
        """Floating IP address"""
        return self.floating_ip_stack.floating_ip_address

    @property
    def server_name(self):
        """Floating IP address"""
        return self.floating_ip_stack.server_name

    @property
    def ssh_client(self):
        """Floating IP address"""
        return self.floating_ip_stack.ssh_client

    def test_ssh(self):
        """Test SSH connectivity to floating IP address"""
        result = sh.execute("hostname", ssh_client=self.ssh_client)
        self.assertEqual([self.server_name.lower()],
                         str(result.stdout).splitlines())

    def test_ssh_from_cli(self):
        """Test SSH connectivity to floating IP address from CLI"""
        result = sh.execute(self.floating_ip_stack.ssh_command + ['hostname'])
        self.assertEqual([self.server_name.lower()],
                         str(result.stdout).splitlines())

    def test_ping(self):
        """Test ICMP connectivity to floating IP address"""
        ping.ping_until_received(self.floating_ip_address).assert_replied()

    # --- test port-security extension ---------------------------------------

    @neutron.skip_if_missing_networking_extensions('port-security')
    def test_port_security_enabled_port_attribute(self):
        """Test port security enabled port attribute"""
        self.assertEqual(self.expected_port_security_enabled,
                         self.observed_port_security_enabled)

    @property
    def expected_port_security_enabled(self):
        """Expected port security enabled value"""
        return self.floating_ip_stack.port_security_enabled

    @property
    def observed_port_security_enabled(self):
        """Actual MTU value for internal network"""
        return self.floating_ip_stack.outputs.port_security_enabled

    # --- test security_group extension --------------------------------------

    @neutron.skip_if_missing_networking_extensions('security-group')
    def test_security_groups_port_attribute(self):
        """Test security groups port attribute"""
        self.assertEqual(self.expected_security_groups,
                         self.observed_security_groups)

    @property
    def expected_security_groups(self):
        """Expected port security groups"""
        return set(self.floating_ip_stack.security_groups)

    @property
    def observed_security_groups(self):
        """Actual port security group"""
        return set(self.floating_ip_stack.outputs.security_groups)

    # --- test net-mtu and net-mtu-writable extensions -----------------------

    @neutron.skip_if_missing_networking_extensions('net-mtu')
    def test_ping_with_net_mtu(self):
        """Test connectivity to floating IP address with MTU sized packets"""
        # Wait until it can reach remote port with maximum-sized packets
        ping.ping(self.floating_ip_address,
                  until=ping.RECEIVED,
                  packet_size=self.observed_net_mtu,
                  fragmentation=False).assert_replied()

        # Verify it can't reach remote port with over-sized packets
        ping.ping(self.floating_ip_address,
                  packet_size=self.observed_net_mtu + 1,
                  fragmentation=False,
                  count=5,
                  check=False).assert_not_replied()

    @neutron.skip_if_missing_networking_extensions('net-mtu-writable')
    def test_mtu_net_attribute(self):
        """Test 'mtu' network attribute"""
        if self.expected_net_mtu:
            self.assertEqual(self.expected_net_mtu,
                             self.observed_net_mtu)

    @property
    def expected_net_mtu(self):
        """Expected MTU value for internal network"""
        return self.floating_ip_stack.internal_network_stack.mtu

    @property
    def observed_net_mtu(self):
        """Actual MTU value for internal network"""
        return self.floating_ip_stack.internal_network_stack.outputs.mtu


@neutron.skip_if_missing_networking_extensions('port-security')
class FloatingIPWithPortSecurityFixture(FloatingIPFixture):
    """Heat stack for testing a floating IP instance with port security enabled

    """

    #: Resources stack with security group to allow ping Nova servers
    security_groups_stack = tobiko.required_setup_fixture(
        _stacks.SecurityGroupsFixture)

    #: Enable port security on internal network
    port_security_enabled = True

    @property
    def security_groups(self):
        """List with ICMP security group"""
        return [self.security_groups_stack.outputs.ssh_security_group_id]


@neutron.skip_if_missing_networking_extensions('port-security',
                                               'security-group')
class FloatingIPWithPortSecurityTest(FloatingIPTest):
    """Tests connectivity to Nova instances via floating IPs with port security

    """

    #: Resources stack with floating IP and Nova server with port security
    floating_ip_stack = tobiko.required_setup_fixture(
        FloatingIPWithPortSecurityFixture)

    def test_ping(self):
        """Test connectivity to floating IP address"""
        # Wait for server instance to get ready by logging in
        tobiko.setup_fixture(self.ssh_client)

        # Check can't reach secured port via floating IP
        ping.ping(self.floating_ip_address,
                  count=5,
                  check=False).assert_not_replied()

    @neutron.skip_if_missing_networking_extensions('net-mtu')
    def test_ping_with_net_mtu(self):
        """Test connectivity to floating IP address"""
        # Wait for server instance to get ready by logging in
        tobiko.setup_fixture(self.ssh_client)

        # Verify it can't reach secured port with maximum-sized packets
        ping.ping(self.floating_ip_address,
                  packet_size=self.observed_net_mtu,
                  fragmentation=False,
                  count=5,
                  check=False).assert_not_replied()

        # Verify it can't reach secured port with over-sized packets
        ping.ping(self.floating_ip_address,
                  packet_size=self.observed_net_mtu + 1,
                  fragmentation=False,
                  count=5,
                  check=False).assert_not_replied()


class FloatingIPWithICMPSecurityGroupFixture(
        FloatingIPWithPortSecurityFixture):
    """Heat stack for testing a floating IP instance with security groups

    """

    @property
    def security_groups(self):
        """List with ICMP security group"""
        return [self.security_groups_stack.outputs.ssh_security_group_id,
                self.security_groups_stack.outputs.icmp_security_group_id]


@neutron.skip_if_missing_networking_extensions('port-security',
                                               'security-group')
class FloatingIPWithICMPSecurityGroupTest(FloatingIPTest):
    """Tests connectivity via floating IP with security ICMP security group

    """
    #: Resources stack with floating IP and Nova server to ping
    floating_ip_stack = tobiko.required_setup_fixture(
        FloatingIPWithICMPSecurityGroupFixture)


class FloatingIPWithNetMtuWritableFixture(FloatingIPFixture):
    """Heat stack for testing setting MTU value to internal network

    """

    #: Heat stack for creating internal network with custom MTU value
    internal_network_stack = tobiko.required_setup_fixture(
        _stacks.InternalNetworkWithNetMtuWritableFixture)


@neutron.skip_if_missing_networking_extensions('net-mtu-writable')
class FlatingIpWithMtuWritableTest(FloatingIPTest):
    """Tests connectivity via floating IP with modified MTU value

    """

    #: Resources stack with floating IP and Nova server
    floating_ip_stack = tobiko.required_setup_fixture(
        FloatingIPWithNetMtuWritableFixture)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections

from oslo_log import log

import tobiko


LOG = log.getLogger(__name__)


def default_keystone_credentials():
    return tobiko.setup_fixture(DefaultKeystoneCredentialsFixture).credentials


class KeystoneCredentials(collections.namedtuple(
        'KeystoneCredentials', ['api_version',
                                'auth_url',
                                'username',
                                'password',
                                'project_name',
                                'domain_name',
                                'user_domain_name',
                                'project_domain_name',
                                'project_domain_id',
                                'trust_id'])):

    def to_dict(self):
        return collections.OrderedDict(
            (k, v)
            for k, v in self._asdict().items()
            if v is not None)

    def __repr__(self):
        params = self.to_dict()
        if 'password' in params:
            params['password'] = '***'
        return 'keystone_credentials({!s})'.format(
            ", ".join("{!s}={!r}".format(k, v)
                      for k, v in params.items()))

    required_params = ('auth_url', 'username', 'password', 'project_name')

    def validate(self, required_params=None):
        required_params = required_params or self.required_params
        missing_params = [p
                          for p in required_params
                          if not getattr(self, p)]
        if missing_params:
            reason = "undefined parameters: {!s}".format(
                ', '.join(missing_params))
            raise InvalidKeystoneCredentials(credentials=self, reason=reason)


def keystone_credentials(api_version=None,
                         auth_url=None,
                         username=None,
                         password=None,
                         project_name=None,
                         domain_name=None,
                         user_domain_name=None,
                         project_domain_name=None,
                         project_domain_id=None,
                         trust_id=None,
                         cls=KeystoneCredentials):
    return cls(api_version=api_version,
               auth_url=auth_url,
               username=username,
               password=password,
               project_name=project_name,
               domain_name=domain_name,
               user_domain_name=user_domain_name,
               project_domain_name=project_domain_name,
               project_domain_id=project_domain_id,
               trust_id=trust_id)


class InvalidKeystoneCredentials(tobiko.TobikoException):
    message = "invalid Keystone credentials; {reason!s}; {credentials!r}"


class EnvironKeystoneCredentialsFixture(tobiko.SharedFixture):

    credentials = None

    def setup_fixture(self):
        from tobiko import config
        auth_url = config.get_env('OS_AUTH_URL')
        if not auth_url:
            LOG.debug("OS_AUTH_URL environment variable not defined")
            return

        api_version = (
            config.get_int_env('OS_IDENTITY_API_VERSION') or
            api_version_from_url(auth_url))
        username = (
            config.get_env('OS_USERNAME') or
            config.get_env('OS_USER_ID'))
        password = config.get_env('OS_PASSWORD')
        project_name = (
            config.get_env('OS_PROJECT_NAME') or
            config.get_env('OS_TENANT_NAME') or
            config.get_env('OS_PROJECT_ID') or
            config.get_env('OS_TENANT_ID'))
        if api_version == 2:
            credentials = keystone_credentials(
                api_version=api_version,
                auth_url=auth_url,
                username=username,
                password=password,
                project_name=project_name)
        else:
            domain_name = (
                config.get_env('OS_DOMAIN_NAME') or
                config.get_env('OS_DOMAIN_ID'))
            user_domain_name = (
                config.get_env('OS_USER_DOMAIN_NAME') or
                config.get_env('OS_USER_DOMAIN_ID'))
            project_domain_name = (
                config.get_env('OS_PROJECT_DOMAIN_NAME'))
            project_domain_id = (
                config.get_env('OS_PROJECT_DOMAIN_ID'))
            trust_id = config.get_env('OS_TRUST_ID')
            credentials = keystone_credentials(
                api_version=api_version,
                auth_url=auth_url,
                username=username,
                password=password,
                project_name=project_name,
                domain_name=domain_name,
                user_domain_name=user_domain_name,
                project_domain_name=project_domain_name,
                project_domain_id=project_domain_id,
                trust_id=trust_id)
        try:
            credentials.validate()
        except InvalidKeystoneCredentials as ex:
            LOG.info("No such valid credentials from environment: %r", ex)
        else:
            self.credentials = credentials


class ConfigKeystoneCredentialsFixture(tobiko.SharedFixture):

    credentials = None

    def setup_fixture(self):
        from tobiko import config
        conf = config.CONF.tobiko.keystone
        auth_url = conf.auth_url
        if not auth_url:
            LOG.debug("auth_url option not defined in 'keystone' section of "
                      "tobiko.conf")
            return

        api_version = (conf.api_version or
                       api_version_from_url(auth_url))
        if api_version == 2:
            credentials = keystone_credentials(
                api_version=api_version,
                auth_url=auth_url,
                username=conf.username,
                password=conf.password,
                project_name=conf.project_name)
        else:
            credentials = keystone_credentials(
                api_version=api_version,
                auth_url=auth_url,
                username=conf.username,
                password=conf.password,
                project_name=conf.project_name,
                domain_name=conf.domain_name,
                user_domain_name=conf.user_domain_name,
                project_domain_name=conf.project_domain_name,
                project_domain_id=conf.project_domain_id,
                trust_id=conf.trust_id)
        try:
            credentials.validate()
        except InvalidKeystoneCredentials as ex:
            LOG.info("No such valid credentials from tobiko.conf: %r", ex)
        else:
            self.credentials = credentials


DEFAULT_KEYSTONE_CREDENTIALS_FIXTURES = [
    EnvironKeystoneCredentialsFixture,
    ConfigKeystoneCredentialsFixture]


class DefaultKeystoneCredentialsFixture(tobiko.SharedFixture):

    fixtures = DEFAULT_KEYSTONE_CREDENTIALS_FIXTURES
    credentials = None

    def setup_fixture(self):
        for fixture in self.fixtures:
            try:
                credentials = tobiko.setup_fixture(fixture).credentials
            except Exception:
                LOG.exception("Error setting up fixture %r", fixture)
            else:
                if credentials:
                    LOG.info("Got default credentials from %r: %r",
                             fixture, credentials)
                    self.credentials = credentials
                    return credentials
        raise RuntimeError('Unable to found any valid credentials')


def api_version_from_url(auth_url):
    if auth_url.endswith('/v2.0'):
        LOG.info('Got Keystone API version 2 from auth_url: %r', auth_url)
        return 2
    elif auth_url.endswith('/v3'):
        LOG.info('Got Keystone API version 3 from auth_url: %r', auth_url)
        return 3
    else:
        LOG.warning('Unable to get Keystone API version from auth_url:  %r',
                    auth_url)
        return None



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import os
import sys

from heatclient.common import template_utils
import yaml

import tobiko


TEMPLATE_SUFFIX = '.yaml'

TEMPLATE_DIRS = list(sys.path)


class HeatTemplate(collections.namedtuple('HeatTemplate',
                                          ['template', 'file', 'files'])):

    _yaml = None

    @classmethod
    def from_dict(cls, template):
        return cls(template=template, file=None, files=None)

    @classmethod
    def from_file(cls, template_file, template_dirs=None):
        if template_dirs or not os.path.isfile(template_file):
            template_dirs = template_dirs or TEMPLATE_DIRS
            template_file = find_heat_template_file(
                template_file=template_file, template_dirs=template_dirs)
        files, template = template_utils.get_template_contents(
            template_file=template_file)
        return cls(file=template_file, files=files, template=template)

    @property
    def yaml(self):
        if not self._yaml:
            self._yaml = yaml.safe_dump(self.template)
        return self._yaml


class HeatTemplateFileFixture(tobiko.SharedFixture):

    template_file = None
    template_dirs = None
    template = None

    def __init__(self, template_file=None, template_dirs=None):
        super(HeatTemplateFileFixture, self).__init__()
        if template_file:
            self.template_file = template_file
        if template_dirs:
            self.template_dirs = template_dirs

    def setup_fixture(self):
        self.setup_template()

    def setup_template(self):
        self.template = HeatTemplate.from_file(
            template_file=self.template_file,
            template_dirs=self.template_dirs)


def heat_template_file(template_file, template_dirs=None):
    return HeatTemplateFileFixture(template_file=template_file,
                                   template_dirs=template_dirs)


def get_heat_template(template_file, template_dirs=None):
    return HeatTemplate.from_file(template_file=template_file,
                                  template_dirs=template_dirs)


def find_heat_template_file(template_file, template_dirs):
    for template_dir in template_dirs:
        template_path = os.path.join(template_dir, template_file)
        if os.path.exists(template_path):
            return template_path

    msg = "Template file {!r} not found in directories {!r}".format(
        template_file, template_dirs)
    raise IOError(msg)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import importlib
import inspect
import weakref
import sys


def load_object(object_id, manager=None, new_loader=None, cached=True):
    manager = manager or LOADERS
    loader = manager.get_loader(object_id=object_id, new_loader=new_loader)
    return loader.load(manager=manager, cached=cached)


def load_module(object_id, manager=None, new_loader=None, cached=True):
    manager = manager or LOADERS
    loader = manager.get_loader(object_id=object_id, new_loader=new_loader)
    return loader.load_module(manager=manager, cached=cached)


class ObjectLoader(object):
    """Previously loaded object meta-data"""

    # Weak reference to target object
    _ref = None

    # Flag that tells if referenced object is an imported module
    _is_module = None

    def __init__(self, object_id):
        # Object ID
        self._id = object_id
        if '.' in object_id:
            # Extract object name and parent id from object id
            parent_id, name = object_id.rsplit('.', 1)
            self._name = name
            self._parent_id = parent_id
        else:
            # Root objects have no parent
            self._name = object_id
            self._parent_id = None

    @property
    def is_module(self):
        return self._is_module

    @property
    def id(self):
        return self._id

    def set(self, obj):
        if obj is None:
            self._is_module = False
            self.get = self._get_none
        elif inspect.ismodule(obj):
            # Cannot create weak reference to modules on Python2
            self._is_module = True
            self.get = self._get_module
        else:
            self._is_module = False
            self._ref = weakref.ref(obj)
            self.get = self._get_object

    def __repr__(self):
        return '{cls!s}({id!r})'.format(cls=type(self).__name__, id=self._id)

    def _get_not_cached(self):
        msg = "Object {!r} not cached".format(self._id)
        raise RuntimeError(msg)

    get = _get_not_cached

    @staticmethod
    def _get_none():
        return None

    def _get_module(self):
        try:
            return sys.modules[self._id]
        except KeyError:
            pass
        return self._get_not_cached()

    def _get_object(self):
        obj = self._ref()
        if obj is None:
            return self._get_not_cached()
        return obj

    def load(self, manager, cached=True):
        if cached:
            try:
                return self.get()
            except RuntimeError:
                pass

        parent_loader = self.get_parent(manager=manager)
        if parent_loader:
            parent = parent_loader.load(manager=manager, cached=cached)
            name = self._name
            try:
                obj = getattr(parent, name)
            except AttributeError:
                if parent_loader.is_module:
                    return self._load_module()
                else:
                    # Child cannot be a module if parent isn't a module
                    raise
            else:
                self.set(obj)
                return obj
        else:
            # Root objects can only be modules
            return self._load_module()

    def _load_module(self):
        obj = importlib.import_module(self._id)
        self.set(obj)
        return obj

    def get_parent(self, manager):
        parent_id = self._parent_id
        if parent_id:
            return manager.get_loader(object_id=parent_id)
        else:
            return None

    def load_module(self, manager, cached=True):
        if cached and self._is_module:
            return self.get()

        self.load(manager=manager, cached=cached)
        if self._is_module:
            return self.get()

        parent = self.get_parent(manager=manager)
        if parent:
            return parent.load_module(manager=manager, cached=True)

        msg = ("Non-module object {!r} has no parent").format(self._id)
        raise RuntimeError(msg)


class LoaderManager(object):

    def __init__(self):
        # Dictionary used to cache object references
        self._loaders = {}

    new_loader = ObjectLoader

    def get_loader(self, object_id, new_loader=None):
        """Get existing ObjectInfo or create new one

        It implements singleton pattern by caching previously created
        ObjectInfo instances to OBJECT_INFOS for later retrieval
        """
        try:
            return self._loaders[object_id]
        except KeyError:
            pass

        new_loader = new_loader or self.new_loader
        loader = new_loader(object_id=object_id)
        if not isinstance(loader, ObjectLoader):
            msg = "{!r} is not instance of class ObjectLoader".format(loader)
            raise TypeError(msg)

        self._loaders[object_id] = loader
        return loader


LOADERS = LoaderManager()
